#include "dfsch-ext/sqlite.h"
#include <dfsch/load.h>
#include <dfsch/hash.h>
#include <dfsch/strings.h>
#include <dfsch/number.h>
#include <sqlite.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

typedef struct sqlite_database_t {
  dfsch_type_t* type;
  sqlite* db;
  int open;
} sqlite_database_t;

static const dfsch_type_t sqlite_database_type = {
  DFSCH_STANDARD_TYPE,
  NULL,
  sizeof(sqlite_database_t),
  "sqlite:database",
  NULL,
  NULL,
  NULL
};

typedef struct sqlite_result_t {
  dfsch_type_t* type;
  sqlite_database_t* db;
  sqlite_vm* vm;
  int open;
  int n_columns;
  char **values;
  char **names;
} sqlite_result_t;

static const dfsch_type_t sqlite_result_type = {
  DFSCH_STANDARD_TYPE,
  NULL,
  sizeof(sqlite_result_t),
  "sqlite:result",
  NULL,
  NULL,
  NULL
};


static void database_finalizer(sqlite_database_t* db, void* cd){
  if (db->open){
    sqlite_close(db->db);
  }
}

static sqlite_database_t* sqlite_database(dfsch_object_t* db){
  if (DFSCH_TYPE_OF(db) != &sqlite_database_type){
    dfsch_error("sqlite:not-a-database", db);
  }

  if (!((sqlite_database_t*)db)->open){
    dfsch_error("sqlite:database-already-closed", db);
  }
  
  return (sqlite_database_t*)db;
}

#define SQLITE_DATABASE_ARG(al, name) \
  DFSCH_GENERIC_ARG(al, name, sqlite_database_t*, sqlite_database)

static void result_finalizer(sqlite_result_t* res, void* cd){
  if (res->open){
    sqlite_finalize(res->vm, NULL);
  }
}

static sqlite_result_t* sqlite_result(dfsch_object_t* res){
  if (DFSCH_TYPE_OF(res) != &sqlite_result_type){
    dfsch_error("sqlite:not-a-result", res);
  }

  if (!((sqlite_result_t*)res)->open){
    dfsch_error("sqlite:result-already-closed", res);
  }
  
  return (sqlite_result_t*)res;
}

#define SQLITE_RESULT_ARG(al, name) \
  DFSCH_GENERIC_ARG(al, name, sqlite_result_t*, sqlite_result)

static dfsch_object_t* native_sqlite_open(void *baton, 
                                          dfsch_object_t* args, 
                                          dfsch_tail_escape_t* esc){
  char* filename;
  char* err;
  dfsch_object_t* creator;
  sqlite_database_t* db;
  int created;
  struct stat s;

  DFSCH_STRING_ARG_OPT(args, filename, ":memory:");
  DFSCH_OBJECT_ARG_OPT(args, creator, NULL);
  DFSCH_ARG_END(args);

  db = (sqlite_database_t*)dfsch_make_object(&sqlite_database_type);
  GC_REGISTER_FINALIZER(db, 
                        (GC_finalization_proc)database_finalizer,
                        NULL, NULL, NULL);
 
  created = (stat(filename, &s)==-1 || !S_ISREG(s.st_mode));
  
  db->db = sqlite_open(filename, 0, &err);

  if (!db->db){
    dfsch_object_t* message;
    message = dfsch_make_string_cstr(err);
    sqlite_freemem(err);
    dfsch_error("sqlite:cannot-open", message);
  }

  db->open = 1;

  if (created && creator){
      dfsch_apply(creator, dfsch_list(1, db));
  }

  return (dfsch_object_t*)db;
}

static dfsch_object_t* native_sqlite_close(void *baton, 
                                           dfsch_object_t* args, 
                                           dfsch_tail_escape_t* esc){
  sqlite_database_t* db;
  SQLITE_DATABASE_ARG(args, db);
  DFSCH_ARG_END(args);

  db->open = 0;
  sqlite_close(db->db);

  return NULL;
}

static dfsch_object_t* native_sqlite_exec(void *baton, 
                                          dfsch_object_t* args, 
                                          dfsch_tail_escape_t* esc){
  sqlite_database_t* db;
  char* sql;
  char* err;
  int ret;

  /*
   * It's relatively good idea to not allow user-defined callbacks here.
   * Propagating exceptions and escape continuations through sqlite code
   * is bad idea and doing it safely is too complicated.
   */

  SQLITE_DATABASE_ARG(args, db);
  DFSCH_STRING_ARG(args, sql);
  DFSCH_ARG_END(args);

  ret = sqlite_exec(db->db, sql, NULL, NULL, &err);
  
  if (ret != SQLITE_OK){
    if (ret == SQLITE_BUSY) {
      dfsch_error("sqlite:busy", (dfsch_object_t*)db);
    }
    dfsch_object_t* message;
    message = dfsch_make_string_cstr(err);
    sqlite_freemem(err);
    dfsch_error("sqlite:error", message);    
  }

  return NULL;
}

static dfsch_object_t* native_sqlite_query(void *baton, 
                                           dfsch_object_t* args, 
                                           dfsch_tail_escape_t* esc){
  sqlite_database_t* db;
  sqlite_result_t* res;
  char* sql;
  char* err;
  int ret;

  SQLITE_DATABASE_ARG(args, db);
  DFSCH_STRING_ARG(args, sql);
  DFSCH_ARG_END(args);
  
  res = (sqlite_result_t*)dfsch_make_object(&sqlite_result_type);

  ret = sqlite_compile(db->db, sql, NULL, &(res->vm), &err);
  if (ret != SQLITE_OK){
    if (ret == SQLITE_BUSY) {
      dfsch_error("sqlite:busy", (dfsch_object_t*)db);
    }
    dfsch_object_t* message;
    message = dfsch_make_string_cstr(err);
    sqlite_freemem(err);
    dfsch_error("sqlite:error", message);    
  }
  res->open = 1;
  
  return (dfsch_object_t*)res;
}

static void finalize_result(sqlite_result_t* res){
  char* err;
  int ret;

  res->open = 0;
  ret = sqlite_finalize(res->vm, &err);
  if (ret != SQLITE_OK){
    dfsch_object_t* message;
    message = dfsch_make_string_cstr(err);
    sqlite_freemem(err);
    dfsch_error("sqlite:error", message);    
  }
}

static dfsch_object_t* native_sqlite_finalize(void *baton, 
                                              dfsch_object_t* args, 
                                              dfsch_tail_escape_t* esc){
  sqlite_result_t* res;

  SQLITE_RESULT_ARG(args, res);
  DFSCH_ARG_END(args);

  finalize_result(res);

  return NULL;
}

static dfsch_object_t* get_row_as_vector(sqlite_result_t* res){
  size_t i;
  dfsch_object_t* vec = dfsch_make_vector(res->n_columns, NULL);

  for (i = 0; i < res->n_columns; i++){
    dfsch_vector_set(vec, i, dfsch_make_string_cstr(res->values[i]));
  }

  return vec;
}
static dfsch_object_t* get_row_as_hash(sqlite_result_t* res){
  size_t i;
  dfsch_object_t* h = dfsch_hash_make(DFSCH_HASH_EQ);

  for (i = 0; i < res->n_columns; i++){
    dfsch_hash_set(h, 
                   dfsch_make_symbol(res->names[i]), 
                   dfsch_make_string_cstr(res->values[i]));
  }

  return h;
}



static dfsch_object_t* native_sqlite_step(void *baton, 
                                          dfsch_object_t* args, 
                                          dfsch_tail_escape_t* esc){
  sqlite_result_t* res;
  char* err;
  int ret;
  dfsch_object_t* format;

  SQLITE_RESULT_ARG(args, res);
  DFSCH_OBJECT_ARG_OPT(args, format, NULL)
  DFSCH_ARG_END(args);


  ret = sqlite_step(res->vm, &res->n_columns, &res->values, &res->names);

  if (ret == SQLITE_ROW){
    if (format == NULL){
      return dfsch_sym_true();
    } else if (dfsch_compare_symbol(format, "vector")) {
      return get_row_as_vector(res);
    } else if (dfsch_compare_symbol(format, "hash")) {
      return get_row_as_hash(res);
    } else {
      dfsch_error("sqlite:unknown-format", format);
    }
  } else if (ret == SQLITE_BUSY) {
    dfsch_error("sqlite:busy", (dfsch_object_t*)res->db);
  } else {
    finalize_result(res);
    return NULL;
  } 
}

static dfsch_object_t* native_sqlite_get_row(void *baton, 
                                             dfsch_object_t* args, 
                                             dfsch_tail_escape_t* esc){
  sqlite_result_t* res;
  dfsch_object_t* format;

  SQLITE_RESULT_ARG(args, res);
  DFSCH_OBJECT_ARG_OPT(args, format, NULL)
  DFSCH_ARG_END(args);

  if (format == NULL || dfsch_compare_symbol(format, "vector")) {
    return get_row_as_vector(res);
  } else if (dfsch_compare_symbol(format, "hash")) {
    return get_row_as_hash(res);
  } else {
    dfsch_error("sqlite:unknown-format", format);
  }
}

static dfsch_object_t* native_sqlite_get_names(void *baton, 
                                               dfsch_object_t* args, 
                                               dfsch_tail_escape_t* esc){
  sqlite_result_t* res;
  size_t i;
  dfsch_object_t* vec;

  SQLITE_RESULT_ARG(args, res);
  DFSCH_ARG_END(args);

  vec = dfsch_make_vector(res->n_columns, NULL);

  for (i = 0; i < res->n_columns; i++){
    dfsch_vector_set(vec, i, dfsch_make_string_cstr(res->names[i]));
  }

  return vec;
}

static dfsch_object_t* native_sqlite_get_types(void *baton, 
                                               dfsch_object_t* args, 
                                               dfsch_tail_escape_t* esc){
  sqlite_result_t* res;
  size_t i;
  dfsch_object_t* vec;

  SQLITE_RESULT_ARG(args, res);
  DFSCH_ARG_END(args);

  vec = dfsch_make_vector(res->n_columns, NULL);

  for (i = 0; i < res->n_columns; i++){
    dfsch_vector_set(vec, i, 
                     dfsch_make_string_cstr(res->names[i + res->n_columns]));
  }

  return vec;
}

static dfsch_object_t* native_sqlite_get_value(void *baton, 
                                               dfsch_object_t* args, 
                                               dfsch_tail_escape_t* esc){
  sqlite_result_t* res;
  size_t column;
  SQLITE_RESULT_ARG(args, res);
  DFSCH_LONG_ARG(args, column);
  DFSCH_ARG_END(args);

  if (column >= res->n_columns){
    dfsch_error("sqlite:column-out-of-bounds", 
                dfsch_make_number_from_long(column));
  }

  return dfsch_make_string_cstr(res->values[column]);
}

static dfsch_object_t* native_sqlite_busy_timeout(void *baton, 
                                                  dfsch_object_t* args, 
                                                  dfsch_tail_escape_t* esc){
  sqlite_database_t* db;
  int timeout;
  SQLITE_DATABASE_ARG(args, db);
  DFSCH_LONG_ARG(args, timeout);
  DFSCH_ARG_END(args);

  sqlite_busy_timeout(db->db, timeout);

  return NULL;
}

static dfsch_object_t* native_sqlite_changes(void *baton, 
                                             dfsch_object_t* args, 
                                             dfsch_tail_escape_t* esc){
  sqlite_database_t* db;
  SQLITE_DATABASE_ARG(args, db);
  DFSCH_ARG_END(args);

  return dfsch_make_number_from_long(sqlite_changes(db->db));
}
static dfsch_object_t* native_sqlite_last_insert_rowid(void *baton, 
                                                       dfsch_object_t* args, 
                                                       dfsch_tail_escape_t* esc){
  sqlite_database_t* db;
  SQLITE_DATABASE_ARG(args, db);
  DFSCH_ARG_END(args);

  return dfsch_make_number_from_long(sqlite_last_insert_rowid(db->db));
}

dfsch_object_t* dfsch_module_sqlite_register(dfsch_object_t* env){
  dfsch_define_cstr(env, "sqlite:<database>", &sqlite_database_type);
  dfsch_define_cstr(env, "sqlite:<result>", &sqlite_result_type);
  

  dfsch_define_cstr(env, "sqlite:open", 
                    dfsch_make_primitive(native_sqlite_open, NULL));
  dfsch_define_cstr(env, "sqlite:close", 
                    dfsch_make_primitive(native_sqlite_close, NULL));
  dfsch_define_cstr(env, "sqlite:exec", 
                    dfsch_make_primitive(native_sqlite_exec, NULL));
  dfsch_define_cstr(env, "sqlite:query", 
                    dfsch_make_primitive(native_sqlite_query, NULL));
  dfsch_define_cstr(env, "sqlite:finalize", 
                    dfsch_make_primitive(native_sqlite_finalize, NULL));
  dfsch_define_cstr(env, "sqlite:step", 
                    dfsch_make_primitive(native_sqlite_step, NULL));
  dfsch_define_cstr(env, "sqlite:get-row", 
                    dfsch_make_primitive(native_sqlite_get_row, NULL));
  dfsch_define_cstr(env, "sqlite:get-names", 
                    dfsch_make_primitive(native_sqlite_get_names, NULL));
  dfsch_define_cstr(env, "sqlite:get-types", 
                    dfsch_make_primitive(native_sqlite_get_types, NULL));
  dfsch_define_cstr(env, "sqlite:get-value", 
                    dfsch_make_primitive(native_sqlite_get_types, NULL));

  dfsch_define_cstr(env, "sqlite:busy-timeout", 
                    dfsch_make_primitive(native_sqlite_busy_timeout, NULL));
  dfsch_define_cstr(env, "sqlite:changes", 
                    dfsch_make_primitive(native_sqlite_changes, NULL));
  dfsch_define_cstr(env, "sqlite:last-insert-rowid", 
                    dfsch_make_primitive(native_sqlite_last_insert_rowid, 
                                         NULL));


  dfsch_provide(env, "sqlite");
  return NULL;
}
